/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;

import java.io.StringReader;

import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.ExpressionFactory;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.javacc.ParseException;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * An <code>ExpressionConfiguration</code> is a fundamental class for parsing
 * and evaluating {@link Expression Expressions}. This class provides the
 * easiest way of working with the <code>de.jtem.mathExpr</code> library and it should be
 * your starting point whenever you decide to work with <code>de.jtem.mathExpr</code>
 * classes.
 * <p>
 * A single instance of this class is a representation of a single
 * <code>Expression</code>.
 * It eases parsing and evaluating by corresponding methods like
 * {@link #setExpression(String exp)} or {@link #evaluateExpression()}.
 * You don't need to create a {@link Parser} or {@link Evaluator} - they were
 * created by this <code>ExpressionConfiguration</code> if necessary.
 * All you have to do is to choose a default {@link Type} (e.g.: {@link
 * de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType#TYPE}) and construct
 * an <code>ExpressionConfiguration</code>.
 * For example:
 * </p><pre>
 * Type defaultType=new ComplexType.TYPE;
 * ExpressionConfiguration config=new ExpressionConfiguration(defaultType);<br>
 * //parse the expression "2+2i"
 * config.{@link #setExpression(String) setExpression}("2+2i");<br>
 * //get the expression as an Expression object
 * Expression exp=config.{@link #getExpression()};<br>
 * //evaluate the expression and get the result as an Object
 * Object o=config.{@link #evaluateExpression()};<br>
 * ...
 * </pre>
 * <p>
 * If you use the constructor {@link #ExpressionConfiguration(Type)}, this
 * class creates a new {@link DefinitionModel} and a new {@link 
 * ExpressionFactory} parameterized with that <code>DefinitionModel</code>.
 * Use the other constructor, if you want to use a single
 * <code>DefinitionModel</code> for several
 * <code>ExpressionConfigurations</code>,
 * e.g.:
 * </p><pre>
 * ExpressionConfiguration config1=new ExpressionConfiguration(ComplexType.TYPE);
 * DefinitionModel defModel=config1.{@link #getDefinitionModel()};
 * ExpressionConfiguration config2=new ExpressionConfiguration(ComplexType.TYPE, defModel);
 * config1.{@link #defineVariable(String,Object,Type) defineVariable}("a", new Complex(2), ComplexType.TYPE);
 * config1.defineVariable("b", new Complex(4), ComplexType.TYPE);
 * config1.{@link #setExpression(String) setExpression}("a+b");
 * //the next command is possible because the DefinitionModel of config1 and config2 is the same instance
 * config2.setExpression("a*b");
 * </pre>
 */
public class ExpressionConfiguration
{
  /**
   * The <code>Expression</code> this <code>ExpressionConfiguration</code>
   * represents.
   */
  Expression expression;
  /**
   * The default <code>Type</code> of evaluating {@link Symbol Symbols}.
   */
  Type defaultEvalType;
  /**
   * The <code>ExpressionFactory</code> any instantiated {@link Parser} needs.
   * 
   * @see #setExpression(String)
   */
  ExpressionFactory expFactory;
  /**
   * The <code>DefinitionModel</code> holding all defined {@link Function
   * Functions} and {@link Variable Variables}.
   * 
   * @see #defineVariable(String,Object,Type)
   * @see #getVariable(String)
   * @see #defineFunction(String,String[],String)
   * @see #defineFunction(String,String[],Expression)
   */
  DefinitionModel defModel;
  /**
   * The <code>Evaluator</code> for this' {@link #Expression}.
   */
  Evaluator evaluator;
  /**
   * Creates a new <code>ExpressionConfiguration</code> instance with
   * {@link ComplexType#TYPE} as the {@link Type} for evaluating
   * {@link Symbol Symbols}.
   */
  public ExpressionConfiguration()
  {
    this(ComplexType.TYPE);
  }
  /**
   * Creates a new <code>ExpressionConfiguration</code> instance with 
   * {@link ComplexType#TYPE} as the {@link Type} for evaluating
   * {@link Symbol Symbols} and
   * the specified <code>DefinitionModel</code>. Use this constructor if you
   * want to have more than one <code>ExpressionConfiguration</code> with
   * the same <code>DefinitionModel</code>.
   * 
   * @param defModel the <code>DefinitionModel</code> to be used for this 
   * <code>ExpressionConfiguration</code>.
   */
  public ExpressionConfiguration(DefinitionModel defModel)
  {
    this(ComplexType.TYPE, defModel);
  }
  /**
   * Creates a new <code>ExpressionConfiguration</code> instance with the
   * specified <code>Type</code> for evaluating {@link Symbol Symbols}.
   * 
   * @param defaultType the <code>Type</code> <code>Symbols</code> will
   * be evaluated to.
   */
  public ExpressionConfiguration(Type defaultType)
  {
    defaultEvalType=defaultType;
    defModel=new DefinitionModel();
    expFactory=new ExpressionFactory(defModel);
  }
  /**
   * Creates a new <code>ExpressionConfiguration</code> instance with the
   * specified <code>Type</code> for evaluating {@link Symbol Symbols} and
   * the specified <code>DefinitionModel</code>. Use this constructor if you
   * want to have more than one <code>ExpressionConfiguration</code> with
   * the same <code>DefinitionModel</code>.
   * 
   * @param defaultType the <code>Type</code> <code>Symbols</code> will
   * be evaluated to.
   * @param defModel the <code>DefinitionModel</code> to be used for this 
   * <code>ExpressionConfiguration</code>.
   */
  public ExpressionConfiguration(Type defaultType, DefinitionModel defModel)
  {
    defaultEvalType=defaultType;
    expFactory=new ExpressionFactory(defModel);
    this.defModel=defModel;
  }
  /**
   * Returns the <code>DefinitionModel</code> holding all {@link Variable
   * Variables} and {@link Function Functions}.
   * 
   * @see #defineVariable(String,Object,Type)
   * @see #defineFunction(String,String[],String)
   * @see #defineFunction(String,String[],Expression)
   * @return the <code>DefinitionModel</code> of this
   * <code>ExpressionConfiguration</code>.
   */
  public DefinitionModel getDefinitionModel()
  {
    return defModel;
  }
  /**
   * <p>Sets a new <code>DefinitionModel</code> for this
   * <code>ExpressionConfiguration</code>.
   * </p>
   * <b>Attention:</b> All of the created definitions are no longer
   * available. Maybe the current setted <code>Expression</code> makes use
   * of some of these definitions. That's why the <code>Expression</code>
   * of this <code>ExpressionConfiguration</code> becomes <code>null</code>
   * if the method <code>setDefinitionModel(DefinitionModel)</code> is
   * calling. You have to reset it.
   *
   * @see #setExpression(String)
   * @see #setExpression(Expression)
   * @see #getDefinitionModel()
   * @param newDefModel the <code>DefinitionModel</code> to be used for this 
   * <code>ExpressionConfiguration</code>.
   */
  public void setDefinitionModel(DefinitionModel newDefModel)
  {
    expression=null;
    evaluator=null;
    expFactory=new ExpressionFactory(defModel);
    defModel=newDefModel;
  }
  /**
   * Returns the <code>ExpressionFactory</code> which is neccessary for any
   * instantiated {@link Parser}. 
   * 
   * @return the <code>ExpressionFactory</code> of this
   * <code>ExpressionConfiguration</code>.
   */
  public ExpressionFactory getExpressionFactory()
  {
    return expFactory;
  }
  /**
   * Returns the current default <code>Type</code> for evaluating {@link Symbol
   * Symbols}.
   *
   * @see #setType(Type)
   * @see #evaluateExpression()
   * @return the <code>Type</code> for evaluating <code>Symbols</code>.
   */
  public Type getType()
  {
    return defaultEvalType;
  }
  /**
   * Sets a new default <code>Type</code> for evaluating {@link Symbol
   * Symbols}.
   *
   * @see #getType()
   * @see #evaluateExpression()
   * @param type the new <code>Type</code> to be set.
   */
  public void setType(Type type)
  {
    defaultEvalType=type;
    evaluator=null;
  }
  /** 
   * Sets the <code>Expression</code> this <code>ExpressionConfiguration</code>
   * is to be representing.
   *
   * @see #setExpression(String)
   * @see #getExpression()
   * @param exp the new <code>Expression</code> for this
   * <code>ExpressionConfiguration</code>.
   */
  public void setExpression(Expression exp)
  {
    if(exp!=expression) 
    {
      expression=exp;
      evaluator=null;
    }
  }
  /** 
   * Sets the <code>Expression</code> this <code>ExpressionConfiguration</code>
   * is to be representing by parsing the specified <code>String</code>. See
   * documentation of class {@link Parser} to get information about the
   * syntax the specified <code>String</code> has to have.<br>
   * <br>
   * You do not need to create a <code>Parser</code>. This method creates a
   * new <code>Parser</code> and sets the <code>Expression</code> of this
   * <code>ExpressionConfiguration</code> object to the parsed one.
   *
   * @see #getEvaluator()
   * @see #setExpression(Expression)
   * @see #getExpression()
   * @param exp a <code>String</code> representing the new
   * <code>Expression</code> for this <code>ExpressionConfiguration</code>.
   */
  public void setExpression(String exp)
  {
    evaluator=null;
    try
    {
      expression=new Parser(new StringReader(exp), expFactory).parse();
    }
    catch(ParseException ex)    
    {
    	throw new IllegalArgumentException( ex.getMessage());
    }  
  }
  /**
   * Returns the <code>Expression</code> this
   * <code>ExpressionConfiguration</code> currently represents or null
   * if an <code>Expression</code> wasn't set yet.
   *
   * @see #setExpression(Expression)
   * @see #setExpression(String)
   * @return the <code>Expression</code> this
   * <code>ExpressionConfiguration</code> currently represents.
   */
  public Expression getExpression()
  {
    return expression;
  }
  /**
   * Returns the current set Expression simplified. Calling this
   * methods delegates to {@link Expression#simplify()}.
   * 
   * @throws IllegalStateException if no Expression was set yet.
   */
  public Expression simplify()
  {
    try
    {
      return expression.simplify();
    }
    catch(NullPointerException ex)
    {
      throw new IllegalStateException("No expression set yet.");
    }
  }
  /**
   * Creates a new Symbol with String str and delegates to {@link
   * #differentiate(Symbol)} and returns the <b>simplified</b> derivative of
   * the current set Expression with respect to that Symbol. 
   * 
   * @see Expression#simplify()
   */
  public Expression differentiate(String str)
  {
    return differentiate(new Symbol(str));
  }
  /**
   * Returns the <b>simplified</b> derivative of the current set Expression with
   * respect to the Symbol s. 
   *
   * @see Expression#differentiate(Symbol)
   * @see Expression#simplify()
   * @return the <b>simplified</b> derivative of the current set Expression with
   * respect to the Symbol s.
   * @throws IllegalStateException if no Expression was set yet.
   */
  public Expression differentiate(Symbol s)
  {
    try
    {
      return expression.differentiate(s);
    }
    catch(NullPointerException ex)
    {
      throw new IllegalStateException("No Expression set yet.");
    }    
  }
  /**
   * Calls {@link Expression#replaceSymbol(Symbol, Expression)} on the current
   * set Expression and returns the result.
   */
  public Expression replaceSymbol(Symbol s, Expression expr)
  {
    return expression.replaceSymbol(s, expr);
  }
  /**
   * Calls {@link Expression#dissolveFunctionCalls()} on the current
   * set Expression and returns the result.
   * 
   * @return a new Expression which is equal to the current set Expression
   * but all function calls are dissolved.
   */
  public Expression dissolveFunctionCalls()
  {
    return expression.dissolveFunctionCalls();
  }
  /**
   * Returns the <code>Evaluator</code> for the current <code>Expression</code>.
   * 
   * @throws NullPointerException if there is not an <code>Expression</code>
   * setted yet.
   * @see #setExpression(Expression)
   * @see #setExpression(String)
   * @return the <code>Evaluator</code> for the current <code>Expression</code>.
   */
  public Evaluator getEvaluator()
  {
    if(expression==null)
      throw new IllegalStateException("Set an expression first.");
    if(evaluator==null)
    {
      try
      {
        evaluator=expression.getEvaluator(defaultEvalType, defModel);
      }
      catch(NullPointerException ex)
      {
        throw new IllegalStateException("Expression required.");
      }
    }
    return evaluator;
  }
  /**
   * Evaluates the current {@link Expression} by calling the method {@link
   * Evaluator#evaluate} on the expression's {@link Evaluator}.
   * <p>
   * This method creates the evaluator-tree for the current expression
   * (the setted expression) if it's not already created (because this
   * method or the method {@link #getEvaluator()} was already called).
   * After that it returns the evaluation result as an Object.
   * </p>
   *
   * @see #getEvaluator()
   * @see #getExpression()
   * @return the <code>Object</code> as the result of the evaluation.
   */
  public Object evaluateExpression() throws UnknownDefinitionException
  {
  	if(evaluator==null)
			evaluator=expression.getEvaluator(defaultEvalType, defModel);
    return evaluator.evaluate();
  }
  /**
   * Defines or changes a <code>Variable</code> by delegating the method {@link
   * DefinitionModel#defineVariable(String,Object,Type)} of this'
   * <code>DefinitionModel</code>. See the documentation of that method for
   * details.
   *
   * @see #getDefinitionModel()
   * @throws IncompatibleDefinitionException if a {@link Function} with the same
   * name is already defined.
   * @return the <code>Variable</code> which was defined respectively changed.
   * @param name the name for the <code>Variable</code> to be defined
   * respectively the name of the <code>Variable</code> to be changed.
   * @param value the value for the <code>Variable</code> to be defined
   * respectively the new value for the <code>Variable</code> to be changed.
   * @param type the <code>Type</code> for the <code>Variable</code> to be
   * defined respectively the new <code>Type</code> for the
   * <code>Variable</code> to be changed.
   */
  public Variable defineVariable(String name, Object value, Type type)
      throws IncompatibleDefinitionException
  {
    return defModel.defineVariable(name, value, type);
  }
  /**
   * Returns the <code>Variable</code> with the specified name by delegating the
   * method {@link DefinitionModel#getVariable(String)} of this'
   * <code>DefinitionModel</code>.
   * 
   * @see #getDefinitionModel()
   * @return the <code>Variable</code> with the specified <code>name</code> or
   * <code>null</code> if the <code>Variable</code> does not exist yet.
   * @param name the name of the <code>Variable</code> this method tries to
   * return.
   */
  public Variable getVariable(String name)
  {
    return defModel.getVariable(name);
  }
  /**
   * Defines or changes a <code>UserDefinedFunction</code> by delegating the
   * method {@link DefinitionModel#defineFunction(String,String[],Expression)}
   * of this' <code>DefinitionModel</code>. See the documentation of that method
   * for details.
   * <br>
   * The specified <code>String def</code> will be parsed before the delegation. 
   *
   * @see #getDefinitionModel()
   * @throws IncompatibleDefinitionException if a {@link Variable} with the same
   * name is already defined.
   * @return the <code>UserDefinedFunction</code> which was defined respectively
   * changed.
   * @param name the name for the <code>UserDefinedFunction</code> to be defined
   * respectively the name of the <code>UserDefinedFunction</code> to be
   * changed.
   * @param param an array of names of the parameters for the
   * <code>UserDefinedFunction</code> to be defined respectively the new
   * parameters for the <code>UserDefinedFunction</code> to be changed.
   * @param def a <code>String</code> representing the definition for the 
   * <code>UserDefinedFunction</code> to be defined respectively a
   * <code>String</code> representing the new definition for the
   * <code>UserDefinedFunction</code> to be changed.
   */
  public UserDefinedFunction defineFunction(String name, String[] param, 
  		String def) //throws IncompatibleDefinitionException
  {
    try
    {
      Expression exp=new Parser(new StringReader(def), expFactory).parse();
      return defModel.defineFunction(name, param, exp);
    }
    catch(ParseException ex)
    {
      System.out.println("ParseException: "+ex.getMessage());
      return null;
    }  
  }
  /**
   * Defines or changes a <code>UserDefinedFunction</code> by delegating the
   * method {@link DefinitionModel#defineFunction(String,String[],Expression)}
   * of this' <code>DefinitionModel</code>. See the documentation of that method
   * for details.
   *
   * @see #getDefinitionModel()
   * @throws IncompatibleDefinitionException if a {@link Variable} with the same
   * name is already defined.
   * @return the <code>UserDefinedFunction</code> which was defined respectively
   * changed.
   * @param name the name for the <code>UserDefinedFunction</code> to be defined
   * respectively the name of the <code>UserDefinedFunction</code> to be
   * changed.
   * @param param an array of names of the parameters for the
   * <code>UserDefinedFunction</code> to be defined respectively the new
   * parameters for the <code>UserDefinedFunction</code> to be changed.
   * @param def the definition for the <code>UserDefinedFunction</code> to be
   * defined respectively the new definition for the <code>UserDefinedFunction
   * </code> to be changed.
   */
  public UserDefinedFunction defineFunction(String name, String[] param, 
  		Expression def) throws IncompatibleDefinitionException
  {
    return defModel.defineFunction(name, param, def);
  }
}
