/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;


import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.parser.Addition;
import de.jtem.mathExpr.parser.Division;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Multiplication;
import de.jtem.mathExpr.parser.NegativeSign;
import de.jtem.mathExpr.parser.Power;
import de.jtem.mathExpr.parser.Symbol;
//for doc-links:


/**
 * A <code>PredefinedFunction</code> is a <code>Function</code> with a
 * predefined definition which can't be redefined (e.g.:
 * <code>sin</code>, <code>cos</code>, <code>log</code>, etc.).
 * Static instances of this <code>PredefinedFunction</code> represent all
 * possible mathematical predefined functions.
 */
public abstract class PredefinedFunction extends Function
{
	/**
	 * This <code>PredefinedFunction</code> instance represents the
	 * <code>sin</code> function.
	 */
  static final public PredefinedFunction SIN =new PredefinedFunction("sin")
  {
    /**
     * Returns the derivative function of sine - that's cosine.
     * paramIndex will be ignored because a sine function
     * has only one parameter.
     * 
     * @return {@link #COS}
     */
    public Function getDerivative(int paramIndex)
    {
      return COS;
    };
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>cos</code> function.
   */
  static final public PredefinedFunction COS =new PredefinedFunction("cos")
  {
    UserDefinedFunction derivative=null;
    /**
     * Returns the derivative function of this cosine - that's -sine.
     * paramIndex will be ignored because a cosine function
     * has only one parameter. The returned Function has the
     * name "Dcos".
     * 
     * @return a new UserDefinedFunction with name "Dcos"
     * and definition "-sin"
     */
    public Function getDerivative(int paramIndex)
    {
      if(derivative==null)
        derivative=new UserDefinedFunction("Dcos", parameter,
            new NegativeSign(new FunctionCall(SIN, toList(new Symbol(parameter[0])))));
      return derivative;
    };
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>cos</code> function.
   */
  static final public PredefinedFunction TAN =new PredefinedFunction("tan")
  {
  	UserDefinedFunction derivative=null;
  	/**
  	 * Returns the derivative function of this cosine - that's -sine.
  	 * paramIndex will be ignored because a cosine function
  	 * has only one parameter. The returned Function has the
  	 * name "Dtan".
  	 * 
  	 * @return a new UserDefinedFunction with name "Dtan"
  	 * and definition "cos(x)^--2"
  	 */
  	public Function getDerivative(int paramIndex)
	{
  		if(derivative==null)
  			derivative=new UserDefinedFunction("Dtan", parameter,
  					new Division( new Symbol( 1 ), new Power(  new FunctionCall(COS, toList(new Symbol(parameter[0]))), new Symbol(2))));
  		return derivative;
  	};
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>exp</code> function.
   */
  static final public PredefinedFunction EXP =new PredefinedFunction("exp")
  {
    /**
     * Returns the derivative function of exponential function - that's exp.
     * paramIndex will be ignored because a exponential function
     * has only one parameter.
     * 
     * @return {@link #EXP}
     */
    public Function getDerivative(int paramIndex)
    {
      return EXP;
    };
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>log</code> function.
   */
  static final public PredefinedFunction LOG =new PredefinedFunction("log")
  {
    UserDefinedFunction derivative=null;
    /**
     * Returns the derivative function of logarithm function - that's 1/param.
     * paramIndex will be ignored because a logarithm function
     * has only one parameter. The returned Function has the
     * name "Dlog".
     * 
     * @return a new UserDefinedFunction with name "Dlog"
     * and definition "1/param"
     */
    public Function getDerivative(int paramIndex)
    {
      if(derivative==null)
        derivative=new UserDefinedFunction("Dlog", parameter,
            new Division(new Symbol(1), new Symbol(parameter[0])));
      return derivative;
    };
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>sqrt</code> function.
   */
  static final public PredefinedFunction SQRT=new PredefinedFunction("sqrt")
  {
    UserDefinedFunction derivative=null;
    /**
     * Returns the derivative function of squareroot function - that's 1/(2*sqrt(param)).
     * paramIndex will be ignored because a logarithm function
     * has only one parameter. The returned Function has the
     * name "Dsqrt".
     * 
     * @return a new UserDefinedFunction with name "Dsqrt"
     * and definition "1/(2*sqrt(param))"
     */
    public Function getDerivative(int paramIndex)
    {
      if(derivative==null)
        derivative=new UserDefinedFunction("Dsqrt", parameter,
            new Division(new Symbol(1), new Multiplication(
            new Symbol(2), new FunctionCall(SQRT, toList(new Symbol(parameter[0]))))));
      return derivative;
    };
  };
  /**
   * This <code>PredefinedFunction</code> instance represents the
   * <code>atan</code> function.
   */
  static final public PredefinedFunction ATAN=new PredefinedFunction("atan")
  {
  	UserDefinedFunction derivative=null;
  	/**
  	 * Returns the derivative function of atan function - that's 1/(1+param^2)).
  	 * paramIndex will be ignored because atan
  	 * has only one parameter. The returned Function has the
  	 * name "Datan".
  	 * 
  	 * @return a new UserDefinedFunction with name "Dsqrt"
  	 * and definition "1/(1+param^2))"
  	 */
  	public Function getDerivative(int paramIndex)
	{
  		if(derivative==null) 
  			derivative=new UserDefinedFunction("Datan", parameter,
  					new Division(new Symbol(1), new Addition( new Symbol(1), new Multiplication(
  							new Symbol(parameter[0]),new Symbol(parameter[0])))));
  		return derivative;
  	};
  };
  /**
   * Stores all <code>static final PredefinedFunction</code> instances which are
   * already available by this class.
   */
  static final public PredefinedFunction[] FUNCTIONS={SIN, COS, TAN, EXP, LOG, SQRT, ATAN};
  /**
   * Instantiates a new <code>PredefinedFunction</code> with the specefied
   * <code>name</code>. This Constructor is protected because all possible
   * <code>PredefinedFunctions</code> are available as <code>static final</code>
   * fields in this class.
   * 
   * @param name the function's name.
   */
  protected PredefinedFunction(String name)
  {
    super(name, new String[] {"p"});//only to define one parameter
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>PredefinedFunction</code>
   * according to the <code>Type</code> the parameter will be evaluated to.
   * Usually the specified <code>Context</code> will be the
   * {@link de.jtem.mathExpr.evaluator.FunctionCallEvaluator} for the
   * {@link de.jtem.mathExpr.parser.FunctionCall} "calling"
   * this <code>PredefinedFunction</code>.<br>
   * This method calls the method 
   * {@link de.jtem.mathExpr.evaluator.Type#getEvaluator(PredefinedFunction,Evaluator)}
   * on the <code>Type</code> the parameter is evaluated to.
   * 
   * @see de.jtem.mathExpr.evaluator.FunctionCallEvaluator#getEvaluator(Symbol)
   * @see de.jtem.mathExpr.evaluator.Evaluator#getReturnType()
   * @return the <code>Evaluator</code> for this
   * <code>PredefinedFunction</code>.
   * @param defaultType not used in this method.
   * @param context the <code>FunctionCallEvaluator</code> for this
   * <code>PredefinedFunction</code> call.
   */
  public Evaluator getCallEvaluator(Type defaultType, Context context)
  {
    Evaluator paramEval=context.getEvaluator(
        new Symbol(getParameter(0)));
    Type paramType=paramEval.getReturnType();
    return paramType.getEvaluator(this, paramEval);
  }
  public Expression insertArguments(Expression param)
  {
    //implementation for one parameter
    return new FunctionCall(this, toList(param));
  }
  private static List toList(Expression expression) {
  	List l=new List();
  	l.addEntry(expression);
  	return l;
  }
}
