/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator.realEvaluator;

import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.evaluator.AbstractType;
import de.jtem.mathExpr.evaluator.Constant;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexPlusRealEvaluator;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexTimesRealEvaluator;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.evaluator.complexEvaluator.RealDivideComplexEvaluator;
import de.jtem.mathExpr.evaluator.complexEvaluator.RealMinusComplexEvaluator;
import de.jtem.mathExpr.evaluator.complexEvaluator.RealPowerComplexEvaluator;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.ComplexVectorType;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.RealTimesComplexVectorEvaluator;
import de.jtem.mathExpr.evaluator.realVectorEvaluator.RealTimesRealVectorEvaluator;
import de.jtem.mathExpr.evaluator.realVectorEvaluator.RealVectorEvaluator;
import de.jtem.mathExpr.evaluator.realVectorEvaluator.RealVectorType;
import de.jtem.mathExpr.parser.BinaryOperation;
//for doc-links:
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.UnaryOperation;

/**
 * The evaluation {@link Type} when a certain part of an {@link Expression}
 * is a {@link Real} when evaluated.
 * The following list shows, what part must be <code>Real</code> so that
 * {@link RealType#TYPE} is used to get the appropriate Evaluator:<br>
 * <ul>
 * <li>{@link BinaryOperation}: the first operand</li>
 * <li>{@link UnaryOperation}: the operand (there is only one)</li>
 * <li>{@link FunctionCall}: nothing (see {@link AbstractType})
 * <li>{@link List}: <b>all</b> entries</li>
 * <li>{@link Symbol}: the appropriate {@link Variable} object must not be of
 * <code>Type RealType</code>. If the <code>Symbol</code> is a constant
 * double value, a new {@link Constant} object of <code>Type RealType</code>
 * will be instantiated. In any other case, the appropriate
 * <code>Variable</code> (or even <code>Constant</code>, if "pi" or something
 * else) will be returned independent from its <code>Type</code>.</li>
 * </ul>
 *
 * @see Symbol#isConstant()
 * @see Constant#getConstant(Symbol,Type)
 * @see Variable#getReturnType()
 */
public class RealType extends AbstractType
{
  /**
   * A <code>static</code> instance of this class. The {@link #RealType()
   * Constructor} of this class is <code>protected</code>, so you have to use
   * this <code>static</code> field.
   */
  static public final RealType TYPE=new RealType();
  /**
   * Creates a new <code>RealType</code> instance. This constructor is
   * <code>protected</code> because you do not need to create several
   * instances of this class. Use the <code>static</code> field {@link
   * #TYPE} instead.
   */
  protected RealType(){}
  /**
   * Returns a {@link Constant} or {@link Variable} represented by the
   * specified <code>Symbol</code>. If the <code>Symbol</code> is a constant
   * value, a new {@link Constant} object of <code>Type RealType</code> will be
   * instantiated. The specified <code>Context</code> contains the
   * <code>Variable</code> if the <code>Symbol</code> is not a
   * constant double value.<br><br>
   * If the <code>Symbol</code> is a mathematical constant like "pi" or "e", the
   * appropriate <code>Constant</code> object will be returned by
   * {@link Constant#getConstant(Symbol,Type)}.
   * 
   * @see Symbol#isConstant()
   * @param context the <code>Context</code>, the <code>Evaluator<code> for
   * the <code>Symbol</code> is getting from (in cases of representing a variable
   * and not a constant).
   */
  public Evaluator getEvaluator(Symbol expr, Context context)
  {
    if(expr.isConstant())
      return new RealConstant(expr.toString(), 
          new Real(expr.getConstantValue()), this);
    Constant c=Constant.getConstant(expr, this);
    if(c!=null)
      return c;
    return context.getEvaluator(expr);
  }
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>UnaryOperation</code> which operand is a {@link Real}
	 * when evaluated. The returned <code>Evaluator</code> will
	 * evaluate to a <code>RealType</code> result again.
	 * 
	 * @see UnaryOperation#getEvaluator(Type,Context)
	 * @see PositiveSignRealEvaluator
	 * @see NegativeSignRealEvaluator
	 * @throws UnsupportedOperationException if the specified
	 * <code>UnaryOperation</code> has an unknown type (see
	 * {@link UnaryOperation#getType()}).
	 */
	public Evaluator getEvaluator(UnaryOperation expr, Evaluator e)
	{
		char uoType=(char)expr.getType();
		switch(uoType)
		{
			case '+': return new PositiveSignRealEvaluator(e);
			case '-': return new NegativeSignRealEvaluator(e);
			default: throw new UnsupportedOperationException((char)expr.getType()+
					": Unknown unary operation.");
		}
	}
  /**
   * Returns the <code>Evaluator</code> for the specified
   * <code>BinaryOperation</code> which first operand is a {@link Real}
   * when evaluated. The returned <code>Evaluator</code> must <b>not</b>
   * evaluate to a <code>RealType</code> result.
   *
   * @see BinaryOperation#getEvaluator(Type,Context)
   * @see RealPlusRealEvaluator
   * @see RealMinusRealEvaluator
   * @see RealTimesRealEvaluator
   * @see RealDivideRealEvaluator
   * @see RealPowerRealEvaluator
   * @see ComplexPlusRealEvaluator
   * @see RealMinusComplexEvaluator
   * @see ComplexTimesRealEvaluator
   * @see RealDivideComplexEvaluator
   * @see RealPowerComplexEvaluator
   * @see RealTimesRealVectorEvaluator
   * @see RealTimesComplexVectorEvaluator
   * @throws UnsupportedOperationException if the specified
	 * <code>BinaryOperation</code> has an unknown type (see
	 * {@link BinaryOperation#getType()}) or a certain binary operation
	 * is not supported (e.g.: Real / RealVector).
   */
  public Evaluator getEvaluator(BinaryOperation expr,
      Evaluator e1, Evaluator e2)
  {
    Type secondOpType=e2.getReturnType();
    char boType=(char)expr.getType();//+, -, * ...
    if(secondOpType==RealType.TYPE)
    {
      switch(boType)
      {
        case '+': return new RealPlusRealEvaluator(e1,e2);
        case '-': return new RealMinusRealEvaluator(e1,e2);
        case '*': return new RealTimesRealEvaluator(e1,e2);
        case '/': return new RealDivideRealEvaluator(e1,e2);
        case '^': return new RealPowerRealEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Real "+boType+
              " Real: Operation is not supported.");
      }
    }
    else if(secondOpType==ComplexType.TYPE)
    {
      switch(boType)
      {
        case '+': return new ComplexPlusRealEvaluator(e2,e1);//commutative
        case '-': return new RealMinusComplexEvaluator(e1,e2);
        case '*': return new ComplexTimesRealEvaluator(e2,e1);//commutative
        case '/': return new RealDivideComplexEvaluator(e1,e2);
        case '^': return new RealPowerComplexEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Real "+boType+
              " Complex: Operation is not supported.");
      }
    }
    else if(secondOpType==RealVectorType.TYPE)
    {
      switch(boType)
      {
        case '*': return new RealTimesRealVectorEvaluator(e1,e2);
        //makes no sense
        //case '/': return new RealDivideRealVectorEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Real "+boType+
              " RealVector: Operation is not supported.");
      }
    }
    else if(secondOpType==ComplexVectorType.TYPE)
    {
      switch(boType)
      {
        case '*': return new RealTimesComplexVectorEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Real "+boType+
              " ComplexVector: Operation is not supported.");
      }
    }
    //else if(secondOpType==RealMatrixType.TYPE)
    //{
    //  switch(boType)
    //  {
    //    default:
    //      throw new UnsupportedOperationException("Real "+boType+
    //          "RealMatrix: Operation is not supported.");
    //  }
    //}
    //else if(secondOpType==ComplexMatrixType.TYPE)
    //{
    //  switch(boType)
    //  {
    //    default:
    //      throw new UnsupportedOperationException("Real "+boType+
    //          "ComplexMatrix: Operation is not supported.");
    //  }
    //}
    throw new UnsupportedOperationException(boType+
        ": Operation is not supported.");
  }
	/**
	 * Returns the <code>Evaluator</code> for the specified <code>List</code>
	 * which all entries are {@link Real} when evaluated. The <code>List</code>
	 * represents a {@link de.jtem.blas.RealVector} in any case and the returned
	 * <code>Evaluator</code> evaluates to a <code>RealVector</code> result.
	 * 
	 * @see List#getEvaluator(Type,Context)
	 * @return a new {@link RealVectorEvaluator} instance
	 */
	public Evaluator getEvaluator(List expr, Evaluator[] e)
	{
    
		return new RealVectorEvaluator(expr, e); 
	}
  /**
   * Returns the <code>Evaluator</code> for the specified
   * <code>PredefinedFunction</code> which parameter is a {@link Real} when
   * evaluated.
   *
   * @see PredefinedFunctionRealEvaluator
   * @throws UnsupportedOperationException if the specified
   * <code>PredefinedFunction</code> is unknown.
   */
  public Evaluator getEvaluator(PredefinedFunction pf, Evaluator paramEval)
  {
    int i;
    for(i=0; i<PredefinedFunction.FUNCTIONS.length; i++)
      if(pf==PredefinedFunction.FUNCTIONS[i])
        break;
    switch(i)
    {
      case 0://sin
        return new PredefinedFunctionRealEvaluator.SinEvaluator(paramEval);
      case 1://cos
        return new PredefinedFunctionRealEvaluator.CosEvaluator(paramEval);
      case 2://tan
      	return new PredefinedFunctionRealEvaluator.TanEvaluator(paramEval);
      case 3://exp
        return new PredefinedFunctionRealEvaluator.ExpEvaluator(paramEval);
      case 4://log
        return new PredefinedFunctionRealEvaluator.LogEvaluator(paramEval);
      case 5://sqrt
        return new PredefinedFunctionRealEvaluator.SqrtEvaluator(paramEval);
      case 6://atan
      	return new PredefinedFunctionRealEvaluator.ATanEvaluator(paramEval);
      default:
        throw new UnsupportedOperationException(pf.getName()+
						": Unknown predefined function.");
    }
  }
  /**
   * Returns {@link Real}<code>.class</code>.
   * 
   * @return <code>Real.class</code>.
   */
  public Class getRepresentationClass()
  {
    return Real.class;
  }
	/**
	 * Returns <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 * 
	 * @param o the <code>Object</code> to compare this <code>RealType</code>
	 * against.
	 * @return <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 */
	public boolean equals(Object o)
	{
		return this==o;
	}
  /**
   * Creates and returns a new {@link RealVariable}.
   * 
   * @see Type#createVariable(String,Object)
   * @return a new <code>RealVariable</code>.
   */
  public Variable createVariable(String name, Object value)
  {
    return new RealVariable(name, (Real)value, this);
  }
}
