/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.example;

import de.jtem.mathExpr.DefinitionModel;
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.UserDefinedFunction;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexEvaluator;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
//for doc-links:
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mfc.field.Complex;

/**
 * A class representing a modifiable <code>ComplexFunction</code>.
 * <p>
 * This class uses an {@link ExpressionConfiguration} holding an {@link
 * Expression} which represents the function's definition. A {@link Variable} of
 * {@link Type} {@link ComplexType#TYPE} will be defined and stored in the
 * {@link DefinitionModel} of the <code>ExpressionConfiguration</code>.
 * That <code>Variable</code> represents the function's parameter.
 * </p><p>
 * Another way to implement this could be using a {@link
 * UserDefinedFunction}.
 * </p>
 */
public class GenericComplexFunction implements Complex.FunctionOnComplex
{
	/**
	 * The <code>ExpressionConfiguration</code> representing this
	 * {@link ComplexFunction ComplexFunction's} definition.
	 */
  ExpressionConfiguration expConfig;
  /**
   * This {@link ComplexFunction ComplexFunction's} parameter.
   */
	Variable zVar;
	/**
	 * The <code>Type</code> {@link Symbol Symbols} in the function's definition
	 * will be evaluated to.
	 */
  Type defaultType;
  /**
   * The <code>String</code> representing this {@link ComplexFunction}.
   */
  String functionStr;
  /**
   * Creates a new <code>GenericComplexFunction</code> with {@link
   * ComplexType#TYPE} as the default evaluation {@link Type}. That
   * <code>Type</code> cannot and must not be changed.
   * <p>
   * Further more this constructor creates a new {@link ExpressionConfiguration}
   * representing the function's definition and defines a new {@link Variable}
   * representing the function's comlex parameter. The function will be
   * initialized to the following: <code>f(z)=z</code>.
   * </p>
   *
   * @see ExpressionConfiguration#defineVariable(String,Object,Type)
   */
  public GenericComplexFunction()
  {
    defaultType=ComplexType.TYPE;
    expConfig=new ExpressionConfiguration(defaultType);
    zVar=expConfig.defineVariable(
        "z", new Complex(0.), ComplexType.TYPE);
    setComplexFunction("z");
  }
  /**
   * Sets the function's defintion by parsing the specified <code>String</code>.
   *
   * @see ExpressionConfiguration#setExpression(String)
   * @param newFunc the <code>String</code> representing the function's
   * definition to be set.
   */
  public void setComplexFunction(String newFunc)
  {
    if(!newFunc.equals(functionStr))
    {
      expConfig.setExpression(newFunc);
      functionStr=newFunc;
    }
  }
  /**
   * Returns the <code>String</code> representing the current function's
   * definition.
   *
   * @return the <code>String</code> representing the current function's
   * definition.
   */
  public String getComplexFunction()
  {
    return functionStr;
  }
  /**
   * Evaluates this <code>ComplexFunction</code> with <code>z</code> as the
   * current parameter value and returns the result as a <b>new</b>
   * <code>Complex</code>.
   *
   * @see ExpressionConfiguration#evaluateExpression()
   * @param z the parameter value for the evaluation.
   * @return a copy of the evaluation result.
   */
  public Complex eval(Complex z)
  {
    zVar.setValue(z);
    return new Complex((Complex)expConfig.evaluateExpression());
  }
  /**
   * Evaluates this <code>ComplexFunction</code> with <code>z</code> as the
   * current parameter value and returns the result in <code>result</code> .
   *
   * @param z the parameter value for the evaluation.
   * @param result the evaluation result will be assigned to this
   * <code>Complex</code>.
   */
  public void eval(Complex z, Complex result)
  {
    zVar.setValue(z);
    ComplexEvaluator e=(ComplexEvaluator)expConfig.getEvaluator();
    e.evaluate();
    e.getComplexValue(result);
  }
}
