/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.example;

import java.awt.Component;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.StringTokenizer;

import javax.swing.JFrame;

import de.jtem.jterm.InterpreterResult;
import de.jtem.jterm.JTerm;
import de.jtem.jterm.Session;
import de.jtem.jterm.StringEvaluator;
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.IncompatibleDefinitionException;
import de.jtem.mathExpr.UnknownDefinitionException;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.evaluator.realEvaluator.RealType;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * An application providing a shell to enter <code>mathExp</code> conform {@link
 * Expression Expressions} or definitions. This application is based on the <a
 * http://www.jtem.de/jterm/index.html>jterm</a> architecture.
 * <p>
 * <h3>Usage:</h3>
 * Type a String representing an <code>Expression</code> or (variable- or
 * function-) definition. To know what kind of syntax an expression
 * <code>String</code> must have, see the documentation of {@link Parser}.
 * A definition <code>String</code> must have the following Syntax:
 * <code>name=expression</code>, e.g.:<br>
 * <pre>
 * &gt;a=5
 * &gt;f(x)=3x
 * </pre>
 * Variables and functions which are defined previously, can be used,
 * e.g.:<br>
 * <pre>
 * &gt;a=5
 * &gt;b=10
 * &gt;f(x)=3x
 * &gt;f(a)*f(b)
 * </pre>
 * If parsing respectively evaluating was successful, it returns a new {@link
 * InterpreterResult} with the <code>String</code> of the created {@link
 * Variable#toString() Variable} or {@link Function#toString() Function} or the
 * result of the evaluated <code>Expression</code>.<br>
 * If an <code>Exception</code> occured during parsing or evaluating, the
 * returned <code>InterpreterResult</code> contains the appropriate error
 * message and {@link InterpreterResult#isErrorMessage}<code>==true</code>.
 * </p>
 * <p>
 * To change the default evaluation {@link Type}, you can type the following:
 * <br>
 * <pre>
 * &gt;defaultType=&lt;type&gt;
 * </pre>
 * For <code>type</code> you can choose <code>real</code> (<code>r</code>) for
 * {@link RealType} or <code>complex</code> (<code>c</code>) for {@link
 * ComplexType}. The initial <code>Type</code> is <code>RealType</code>.
 */
public class MathExprJTerm
{
  public Session session;
  public static void main(String[] args)
  {
    new MathExprJTerm();
  }
  /**
   * Creates a new <code>MathExprJTerm</code> and the application's GUI.
   */
  public MathExprJTerm()
  {
    StringEvaluator bsh = new MathExpStringEvaluator();
    final JFrame top = new JFrame("JTerm for de.jtem.mathExpr");
    session = new Session(bsh);
    final JTerm frame = new JTerm(session);
    top.setContentPane(frame);
    top.addWindowListener(new WindowAdapter()
    {
      Component[] c=frame.getComponents();
      public void windowClosing(WindowEvent e)
      {
        System.exit(0);
      }
      public void windowActivated(WindowEvent e)
      {
        frame.requestFocus();
      }
    });
    top.pack();
    top.setVisible(true);
  }
}
//=============================================================================
/**
 * The <code>StringEvaluator</code> implementation for parsing and evaluating
 * {@link Expression Expressions} regarding to the <code>mathExp</code> library.
 */
class MathExpStringEvaluator implements StringEvaluator
{
  private Type defaultEvalType=RealType.TYPE;
  private ExpressionConfiguration config=
      new ExpressionConfiguration(defaultEvalType);
  /**
   * Parses every {@link Expression} or definition <code>String</code> and
   * returns an appropriate <code>InterpreterResult</code>.
   * 
   * @param script the expression- or definition-<code>String</code> to be
   * parsed.
   * @return the InterpreterResult holding the evaluation result, new definition
   * or error message.
   */
  public InterpreterResult evaluate(String script)
  {
    int defNameLength=0;
    int index=script.indexOf('=');
    //expression or definition?
    if(index==-1)
    {
      //expression
      try
      {
        config.setExpression(script);
        return new InterpreterResult(config.evaluateExpression().toString());
      }
      catch(IllegalArgumentException ex)
      {
        return new InterpreterResult(ex.getMessage(), true);
      }
      catch(UnknownDefinitionException ex)
      {
        return new InterpreterResult(ex.getMessage(), true);
      }
      catch(UnsupportedOperationException ex)
      {
        return new InterpreterResult(ex.getMessage(), true);
      }
    }
    else
    {
      //definition or DefaultTypeChange
      return parseDefinition(index, script);
    }
  }
  /**
   * Parses the specified <code>String</code> representing a variable- or
   * function-definition and returns a new <code>InterpreterResult</code>
   * with the <code>String</code> of the created {@link Variable#toString()
   * Variable} or {@link Function#toString() Function}.
   * If an <code>Exception</code> occured during parsing or evaluating, the
   * returned <code>InterpreterResult</code> contains the appropriate error
   * message.
   * 
   * @see InterpreterResult#isErrorMessage
   */
  private InterpreterResult parseDefinition(int index, String script)
  {
    String defNameStr=script.substring(0, index);
    String definition=script.substring(index+1);
    //variable- or function-definition?
    int leftBracketIndex=defNameStr.indexOf('(');
    int rightBracketIndex=defNameStr.indexOf(')');
    if(leftBracketIndex==-1)
    {
      //variable-definition or DefaultTypeChange
      Variable v;
      defNameStr.trim();
      definition.trim();
      if(defNameStr.equalsIgnoreCase("defaultType"))
      {
        //DefaultTypeChange
        if(definition.equalsIgnoreCase("r") || 
            definition.equalsIgnoreCase("real"))
        {
          defaultEvalType=RealType.TYPE;
          config.setType(defaultEvalType);
          return new InterpreterResult("default Type changed to RealType");
        }
        else if(definition.equalsIgnoreCase("c") || 
            definition.equalsIgnoreCase("complex"))
        {
          defaultEvalType=ComplexType.TYPE;
          config.setType(defaultEvalType);
          return new InterpreterResult("default Type changed to ComplexType");
        }
        else
          return new InterpreterResult(definition+": Type not supported. "+
              "Only \"real\" or \"complex\" are possible.", true);
      }
      else
      {
        //variable-definition
        try
        {
          config.setExpression(definition);
        }
        catch(IllegalArgumentException ex)
        {
          return new InterpreterResult(ex.getMessage(), true);
        }
        catch(UnsupportedOperationException ex)
        {
          return new InterpreterResult(ex.getMessage(), true);
        }
        Object result=config.evaluateExpression();
        Type resultType=config.getEvaluator().getReturnType();
        //System.out.println("Type: "+resultType);
        try
        {
          v=config.defineVariable(defNameStr, result, resultType);
        }
        catch(IncompatibleDefinitionException ex)
        {
          return new InterpreterResult("Variable \""+defNameStr+
              "\" can't be defined"+" because there is already a Function \""+
              defNameStr+"\".", true);
        }
        catch(UnsupportedOperationException ex)
        {
          return new InterpreterResult(ex.getMessage(), true);
        }
        return new InterpreterResult(v.toString());
      }
    }
    else
    {
      //function-definition
      String funcNameStr=defNameStr.substring(0, leftBracketIndex).trim();
      String parameterStr=defNameStr.
          substring(leftBracketIndex+1, rightBracketIndex).trim();
      int nextComma=0;
      int offset=0;
      int commas=0;
      while(nextComma!=-1)//how many parameters
      {
        nextComma=parameterStr.indexOf(',', offset);
        if(nextComma!=-1)
          commas++;
        offset=nextComma+1;
      }
      String[] paramNames=new String[commas+1];
      StringTokenizer tokenizer=new StringTokenizer(parameterStr, ",");
      int i=0;
      while(tokenizer.hasMoreTokens())
      {
        paramNames[i]=tokenizer.nextToken().trim();
        i++;
      }
      Function f;
      try
      {
        f=config.defineFunction(funcNameStr, paramNames, definition);
      }
      catch(IllegalArgumentException ex)
      {
        return new InterpreterResult(ex.getMessage(), true);
      }
      catch(UnsupportedOperationException ex)
      {
        return new InterpreterResult(ex.getMessage(), true);
      }
      catch(IncompatibleDefinitionException ex)
      {
        return new InterpreterResult("Function \""+funcNameStr+
            "\" can't be defined"+" because there is already a Variable \""+
            funcNameStr+"\".", true);
      }
      return new InterpreterResult(f.toString());
    }
  }
  /**
   * For this {@link StringEvaluator} it only returns a <code>String</code>
   * array with only one empty <code>String</code> because there will be
   * never something to complete.
   */
  public String[] completeCommand(String part)
  {
    String[] nothing={""};
    return nothing;
  }
}
