/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import java.util.Collection;
import java.util.HashSet;

//for doc-links:
import de.jtem.mathExpr.DefinitionModel;
import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.UserDefinedFunction;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.FunctionCallEvaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * An <code>Expression</code> describes a mathematical expression (term)
 * and usually it is the result of a <code>String</code>-parsing by
 * the {@link Parser}.
 * <p>
 * Every <code>Expression</code> consists in further <code>Expressions</code>
 * building an expression-tree-hirarchy. <code>BinaryOperations</code>,
 * <code>UnaryOperations</code> and <code>Lists</code> are containers holding
 * at least one other <code>Expression</code>. <code>Symbols</code> are the
 * atomic components in the expression-tree-hirarchy - they are the "leafs of
 * the tree".
 * </p>
 *
 * @see Parser#parse()
 */
public abstract class Expression implements Cloneable
{
  /**
   * Return the evaluation priority of this <code>Expression</code>. For
   * example, a {@link BinaryOperation} of type '+' returns a lower priority
   * than a <code>BinaryOperation</code> of type '*' (3+4*5).
   * <p>
   * The priority of any <code>Expression</code> decides on the structure 
   * in the expression-tree-hirarchy respectively the correct sequence of the
   * later <code>Evaluator</code> tree.
   * </p>
   * <p>
   * Usually the <code>Parser</code> respectively the {@link ExpressionFactory}
   * defines the priority of any created <code>Expression</code>.
   * </p>
   *
   * @see Parser#parse()
   * @see #getEvaluator(Type,Context)
   * @return the evaluation priority of this <code>Expression</code>.
   */
  protected abstract int getPriority();
  /**
   * Returns <code>true</code> if this <code>Expression</code> is a
   * constant value, false otherwise.
   * If this <code>Expression</code> is not a <code>Symbol</code>,
   * it's constant if all Sub-<code>Expressions</code> (e.g.: the operands
   * of a <code>BinaryOperation</code>) are constant in a kind
   * of recursion.
   *
   * @return <code>true</code> if this <code>Expression</code> is a
   * constant value, false otherwise.
   */
  public abstract boolean isConstant();
  /**
   * Return <code>true</code> if this <code>Expressions</code> is ready
   * to get evaluated, false otherwise.
   * If this <code>Expression</code> is not a <code>Symbol</code>,
   * it's evaluatable if all Sub-<code>Expressions</code> (e.g.: the operands
   * of a <code>BinaryOperation</code>) are evaluatable in a kind
   * of recursion.
   * 
   * @return <code>true</code> if this <code>Expressions</code> is ready
   * to get evaluated, false otherwise.
   */
  public abstract boolean isComplete();
  /**
   * Returns a <code>Collection</code> of the names of all
   * {@link Variable Variables} which are essential to evaluate this
   * <code>Expression</code>. In other words: If at least one
   * <code>Variable</code> (the returning <code>Collection</code> contains) is
   * not defined by the {@link DefinitionModel} yet, this
   * <code>Expression</code> can't be evaluated.<br>
   * Use {@link #getEssentialVariables(Collection)}
   * to store the <code>Variable</code> names in a Collection defined by
   * yourself.
   * <p>
   * <b>Attention:</b> If you call this method on an <code>Expression</code>
   * which is a {@link UserDefinedFunction} definition, {@link Symbol Symbols}
   * representing function parameters in that
   * <code>Expression</code> can't be differ from <code>Symbols</code>
   * representing global <code>Variables</code>. That's why the parameter names
   * will be also added to the <code>Collection</code>, although they are not
   * representing essential <code>Variables</code> because
   * <code>Evaluators</code> for the
   * parameters will be provided by the appropriate {@link
   * FunctionCallEvaluator} and not by <code>Variable</code> objects.
   * </p>
   *
   * @see UserDefinedFunction#getDefinition()
   * @see FunctionCallEvaluator#getEvaluator(Symbol)
   * @return a Collection of <code>Variable</code> names
   */
  public Collection getEssentialVariables()
  {
    Collection c=new HashSet();
    getEssentialVariables(c);
    return c;
  }
  /**
   * Add all names of essential <code>Variables</code> into the specified
   * <code>Collection</code>. See {@link #getEssentialVariables()} for details.
   *
   * @param c the <code>Collection</code> the <code>Variable</code> names
   * will be added to.
   */
  public abstract void getEssentialVariables(Collection c);
  /**
   * Return the <code>Evaluator</code> for this <code>Expression</code>.
   * The specified <code>Type</code> determines the kind the ReturnType of
   * any {@link Symbol} in the expression-tree-hirarchy should be evaluated to.
   * The specified <code>Context</code> provides the evaluation of that
   * <code>Symbols</code>, which can be global <code>Variables</code> but
   * also function-parameters.
   *
   * @see Evaluator#getReturnType()
   * @see Context#getEvaluator(Symbol)
   * @return the <code>Evaluator</code> for this <code>Expression</code>.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluators<code> for
   * the <code>Symbols</code> are getting from.
   */
  public abstract Evaluator getEvaluator(Type defaultType,
      Context context);
  /**
   * Returns a new Expression which is equal to this but simplified.
   * <p>
   * The following table shows source operations and it's simplifications. a and b are
   * arbitrary expressions. Note that often there is no simplification (e.g. 1 + a).</p>
   * <table border=1 CELLPADDING="3" CELLSPACING="0">
   * <tr><th>source operation #</th><th>+</th><th>-</th><th>*</th><th>/</th><th>^</th></tr>
   * <tr><th>#a</th><td>a</td><td><font color=#AAAAAA>-a</font></td><td rowspan="2" colspan="3" align="center">n.a.</td></tr>
   * <tr><th>#(#a)</th><td>a</td><td>a</td></tr>
   * <tr><th>0 # a</th><td>a</td><td>-a</td><td>0</td><td>0</td><td>0</td></tr>
   * <tr><th>a # 0</th><td>a</td><td>a</td><td>0</td><td><font color=#AAAAAA>a / 0</font></td><td>1</td></tr>
   * <tr><th>1 # a</th><td><font color=#AAAAAA>1 + a</font></td><td><font color=#AAAAAA>1 - a</font></td><td>a</td><td><font color=#AAAAAA>1 / a</font></td><td>1</td></tr>
   * <tr><th>a # 1</th><td><font color=#AAAAAA>a + 1</font></td><td><font color=#AAAAAA>a - 1</font></td><td>a</td><td>a</td><td>a</td></tr>
   * <tr><th>a # a</th><td><font color=#AAAAAA>a + a</font></td><td>0</td><td><font color=#AAAAAA>a * a</font></td><td>1</td><td><font color=#AAAAAA>a ^ a</font></td></tr>
   * <tr><th>a # b</th><td><font color=#AAAAAA>a + b</font></td><td><font color=#AAAAAA>a - b</font></td><td><font color=#AAAAAA>a * b</font></td><td><font color=#AAAAAA>a / b</font></td><td><font color=#AAAAAA>a ^ b</font></td></tr>
   * <tr><th>-a # b</th><td>b - a</td><td>-(a + b)</td><td>-(a * b)</td><td>-(a / b)</td><td><font color=#AAAAAA>-a ^ b</font></td></tr>
   * <tr><th>a # -b</th><td>a - b</td><td>a + b</td><td>-(a * b)</td><td>-(a / b)</td><td><font color=#AAAAAA>a ^ -b</font></td></tr>
   * <tr><th>-a # -b</th><td>-(a + b)</td><td>b - a</td><td>a * b</td><td>a / b</td><td><font color=#AAAAAA>-(a ^ -b)</font></td></tr>
   * </table>
   * <p>
   * Note: "-0" will be seen equal to "0" in any case (e.g.: "1/-0" -> "1/0")</p>
   *
   * @return a new Expression which is equal to this but simplified.
   */
  public abstract Expression simplify();
  /**
   * Returns the derivative of this Expression with respect to the Symbol s.
   * This method should also {@link #simplify() simplify} the derivative as
   * well as possible before returning it.
   *
   * @return a new Expression which is the derivative of this with respect to s.
   */
  public abstract Expression differentiate(Symbol s);
  /**
   * Creates a new Symbol with String str and delegates to {@link
   * #differentiate(Symbol)}.
   */
  public Expression differentiate(String str)
  {
    return differentiate(new Symbol(str));
  }
  /**
   * Replaces all Symbols equal to s in the expression-tree-hierarchy
   * with the specified Expression. Note that this method returns
   * a new Expression object - this Expression will not be changed.
   * 
   * @see Symbol#equals(Object)
   * @param s the Symbol to be replace.
   * @param expr the Expression for replacement. 
   * @return a new Expression as a result of replacing
   * any s of this Expression with expr.
   */
  public abstract Expression replaceSymbol(Symbol s, Expression expr);
  /**
   * Inserts the arguments of any FunctionCall in the expression-tree-hierarchy
   * into the appropriate function definitions and replaces the FunctionCalls
   * by that new Expressions. This method returns a new Expression
   * object - this Expression does not change.
   * 
   * @see Function#insertArguments(Expression)
   * @return A new Expression differing from this in the way that all
   * FunctionCalls are dissolved. The returned expression-tree-hierarchy
   * does not contain any FunctionCalls.
   */
  public abstract Expression dissolveFunctionCalls();
}
